makesis rexx_marm.pkg
del rexx.sis
ren rexx_marm.sis rexx.sis
makesis rexx_wins.pkg
del rexxW.sis
ren rexx_wins.sis rexxW.sis
makesis rexxstart_marm.pkg
del rexxstart.sis
ren rexxstart_marm.sis rexxstart.sis
makesis rexxstart_wins.pkg
del rexxstartW.sis
ren rexxstart_wins.sis rexxstartW.sis
